#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#if defined(WIN32)
        #include <io.h>
#else
        #include <unistd.h>
		#include "portable.h"
#endif


typedef struct {

	int           pid;
	int           last_continuity_counter;

	__int64       total;
	__int64       drop;
	__int64       scrambling;

	unsigned char last_packet[188];
	int           duplicate_count;

} TS_STATUS;

typedef struct {
	int           sync;
	int           transport_error_indicator;
	int           payload_unit_start_indicator;
	int           transport_priority;
	int           pid;
	int           transport_scrambling_control;
	int           adaptation_field_control;
	int           continuity_counter;
} TS_HEADER;

static void show_usage();

static void tsdump(const char *path);
static void tsselect(const char *src, const char *dst, const char *pid);
static int select_unit_size(unsigned char *head, unsigned char *tail);
static unsigned char *resync(unsigned char *head, unsigned char *tail, int unit_size);
static unsigned char *resync_force(unsigned char *head, unsigned char *tail, int unit_size);
static void extract_ts_header(TS_HEADER *dst, unsigned char *packet);
static int check_unit_invert(unsigned char *head, unsigned char *tail);

int main(int argc, char **argv)
{
	if(argc < 2){
		show_usage();
		exit(EXIT_FAILURE);
	}

	if(argc == 2){
		tsdump(argv[1]);
	}else{
		int i,n;
		unsigned char pid[8192];

		memset(pid, 0, sizeof(pid));
		for(i=3;i<argc;i++){
			n = strtol(argv[i], NULL, 0);
			if( (n >= 0) || (n < 8192) ){
				pid[n] = 1;
			}
		}

		tsselect(argv[1], argv[2], (char *)pid);

	}

	return EXIT_SUCCESS;
}

static void show_usage()
{
	fprintf(stderr, "tsselect - MPEG-2 TS stream(pid) selector ver. 0.1.4\n");
	fprintf(stderr, "usage: tsselect src.m2t [dst.m2t pid  [more pid ..]]\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "ex: dump \"src.m2t\" TS information\n");
	fprintf(stderr, "  tsselect src.m2t\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "ex: remux \"src.m2t\" to \"dst.m2t\" which contains pid=0x1000 and pid=0x1001\n");
	fprintf(stderr, "  tsselect src.m2t dst.m2t 0x1000 0x1001\n");
	fprintf(stderr, "\n");
}

static void tsdump(const char *path)
{
	int fd;

	int pid;
	int lcc;
	int i,m,n;
	int max_buff;
	int unit_size;

	TS_STATUS *stat;
	TS_HEADER  header;

	unsigned char *p;
	unsigned char *curr;
	unsigned char *tail;

	unsigned char buf[8192];

	fd = -1;
	stat = NULL;

	fd = _open(path, _O_BINARY|_O_RDONLY|_O_SEQUENTIAL);
	if(fd < 0){
		fprintf(stderr, "error - failed on open(%s) [src]\n", path);
		goto LAST;
	}

	stat = (TS_STATUS *)calloc(8192, sizeof(TS_STATUS));
	if(stat == NULL){
		fprintf(stderr, "error - failed on malloc(size=%zd)\n", sizeof(TS_STATUS)*8192);
		goto LAST;
	}

	for(i=0;i<8192;i++){
		stat[i].pid = i;
		stat[i].last_continuity_counter = -1;
	}

	n = _read(fd, buf, sizeof(buf));

	unit_size = select_unit_size(buf, buf+n);
	if(unit_size < 188){
		fprintf(stderr, "error - failed on select_unit_size()\n");
		goto LAST;
	}

	do {
		curr = buf;
		tail = buf + n;
		while( (curr+unit_size) <= tail ){
			if(curr[0] != 0x47){
				p = resync(curr, tail, unit_size);
				if(p == NULL){
					break;
				}
				curr = p;
				if( (curr+unit_size) > tail ){
					break;
				}
			}
			extract_ts_header(&header, curr);
			pid = header.pid;
			lcc = stat[pid].last_continuity_counter;
			if(lcc >= 0){
				if( pid == 0x1fff ){
					// null packet - drop count has no mean
					// do nothing
				}else if( (header.adaptation_field_control & 0x01) == 0 ){
					// no payload : continuity_counter should not increment
					if(lcc != header.continuity_counter){
						stat[pid].drop += 1;
					}
				}else if(lcc == header.continuity_counter){
					// has payload and same continuity_counter
					if(memcmp(stat[pid].last_packet, curr, 188) != 0){
						// non-duplicate packet
						stat[pid].drop += 1;
					}
					stat[pid].duplicate_count += 1;
					if(stat[pid].duplicate_count > 1){
						// duplicate packet count exceeds limit (two)
						stat[pid].drop += 1;
					}
				}else{
					m = (lcc + 1) & 0x0f;
					if(m != header.continuity_counter){
						stat[pid].drop += 1;
					}
					stat[pid].duplicate_count = 0;
				}
			}
			stat[pid].last_continuity_counter = header.continuity_counter;
			stat[pid].total += 1;
			memcpy(stat[pid].last_packet, curr, 188);
			if(header.transport_scrambling_control){
				stat[pid].scrambling += 1;
			}
			curr += unit_size;
		}

		n = tail - curr;
		if(n > 0){
			memcpy(buf, curr, n);
		}
		m = _read(fd, buf+n, sizeof(buf)-n);
		if(m < 1){
			break;
		}
		n += m;

	} while(n > unit_size);

	curr = buf;
	tail = buf + n;
	while( (curr+188) <= tail ){
		if(curr[0] != 0x47){
			p = resync_force(curr, tail, unit_size);
			if(p == NULL){
				break;
			}
			curr = p;
			if( (p+188) > tail ){
				break;
			}
		}
		extract_ts_header(&header, curr);
		pid = header.pid;
		lcc = stat[pid].last_continuity_counter;
		if(lcc >= 0){
			if( pid == 0x1fff ){
				// null packet - drop count has no mean
				// do nothing
			}else if( (header.adaptation_field_control & 0x01) == 0 ){
				// no payload : continuity_counter should not increment
				if(lcc != header.continuity_counter){
					stat[pid].drop += 1;
				}
			}else if(lcc == header.continuity_counter){
				// has payload and same continuity_counter
				if(memcmp(stat[pid].last_packet, curr, 188) != 0){
					// non-duplicate packet
					stat[pid].drop += 1;
				}
				stat[pid].duplicate_count += 1;
				if(stat[pid].duplicate_count > 1){
					// duplicate packet count exceeds limit (two)
					stat[pid].drop += 1;
				}
			}else{
				m = (lcc + 1) & 0x0f;
				if(m != header.continuity_counter){
					stat[pid].drop += 1;
				}
				stat[pid].duplicate_count = 0;
			}
		}
		stat[pid].last_continuity_counter = header.continuity_counter;
		stat[pid].total += 1;
		memcpy(stat[pid].last_packet, curr, 188);
		if(header.transport_scrambling_control){
			stat[pid].scrambling += 1;
		}
		curr += unit_size;
	}

LAST:
	if(stat){
		for(i=0;i<8192;i++){
			if(stat[i].total > 0){
				printf("pid=0x%04x, total=%8" PRId64 ", drop=%3" PRId64  ", scrambling=%" PRId64 "\n", i, stat[i].total, stat[i].drop, stat[i].scrambling);
			}
		}
		free(stat);
		stat = NULL;
	}

	if(fd >= 0){
		_close(fd);
		fd = -1;
	}
}

static void tsselect(const char *src, const char *dst, const char *pid)
{
	int sfd,dfd;

	int m,n;
	int unit_size;

	TS_HEADER header;

	unsigned char *p;
	unsigned char *curr;
	unsigned char *tail;

	unsigned char buf[8192];

	sfd = -1;
	dfd = -1;

	sfd = _open(src, _O_BINARY|_O_RDONLY|_O_SEQUENTIAL);
	if(sfd < 0){
		fprintf(stderr, "error - failed on open(%s) [src]\n", src);
		goto LAST;
	}

	dfd = _open(dst, _O_WRONLY|_O_BINARY|_O_CREAT|_O_TRUNC, _S_IREAD|_S_IWRITE);
	if(dfd < 0){
		fprintf(stderr, "error - failed on open(%s) [dst]\n", dst);
		goto LAST;
	}

	n = _read(sfd, buf, sizeof(buf));

	unit_size = select_unit_size(buf, buf+n);
	if(unit_size < 188){
		fprintf(stderr, "error - failed on select_unit_size()\n");
		goto LAST;
	}

	do {
		curr = buf;
		tail = buf + n;
		while( (curr+unit_size) <= tail ){
			if(curr[0] != 0x47){
				p = resync(curr, tail, unit_size);
				if(p == NULL){
					break;
				}
				curr = p;
				if( (curr+unit_size) > tail ){
					break;
				}
			}
			extract_ts_header(&header, curr);
			if(pid[header.pid] != 0){
				m = _write(dfd, curr, 188);
				if(m != 188){
					fprintf(stderr, "error - failed on write() [dst]\n");
					goto LAST;
				}
			}
			curr += unit_size;
		}

		n = tail - curr;
		if(n > 0){
			memcpy(buf, curr, n);
		}
		m = _read(sfd, buf+n, sizeof(buf)-n);
		if(m < 1){
			break;
		}
		n += m;
	}while(n > unit_size);

	curr = buf;
	tail = buf + n;
	while( (curr+188) <= tail ){
		if(curr[0] != 0x47){
			p = resync_force(curr, tail, unit_size);
			if(p == NULL){
				break;
			}
			curr = p;
			if( (p+188) > tail ){
				break;
			}
		}
		extract_ts_header(&header, curr);
		if(pid[header.pid] != 0){
			m = _write(dfd, curr, 188);
			if(m != 188){
				fprintf(stderr, "error - failed on write() [dst]\n");
				goto LAST;
			}
		}
		curr += unit_size;
	}

LAST:
	if(dfd >= 0){
		_close(dfd);
		dfd = -1;
	}
	if(sfd >= 0){
		_close(sfd);
		sfd = -1;
	}
}

static int select_unit_size(unsigned char *head, unsigned char *tail)
{
	int i;
	int m,n,w;
	int count[320];

	unsigned char *pre;
	unsigned char *buf;

	pre = NULL;
	buf = head;
	memset(count, 0, sizeof(count));

	// 1st step, find head 0x47
	while(buf < tail){
		if(buf[0] == 0x47){
			pre = buf;
			break;
		}
		buf += 1;
	}

	if(pre == NULL){
		return 0;
	}

	// 2nd step, count up 0x47 interval
	buf = pre + 1;
	while( buf < tail ){
		if(buf[0] == 0x47){
			m = buf - pre;
			if(m < 188){
				n = check_unit_invert(head, buf);
				if( (n >= 188) && (n < 320) ){
					count[n] += 1;
					pre = buf;
				}
			}else if(m < 320){
				count[m] += 1;
				pre = buf;
			}else{
				pre = buf;
			}
		}
		buf += 1;
	}

	// 3rd step, select maximum appeared interval
	m = 0;
	n = 0;
	for(i=188;i<320;i++){
		if(m < count[i]){
			m = count[i];
			n = i;
		}
	}

	// 4th step, verify unit_size
	w = m*n;
	if( (m < 8) || (w < ((tail-head) - (w/8))) ){
		return 0;
	}

	return n;
}

static unsigned char *resync(unsigned char *head, unsigned char *tail, int unit_size)
{
	int i;
	unsigned char *buf;

	buf = head;
	tail -= unit_size * 8;
	while( buf < tail ){
		if(buf[0] == 0x47){
			for(i=1;i<8;i++){
				if(buf[unit_size*i] != 0x47){
					break;
				}
			}
			if(i == 8){
				return buf;
			}
		}
		buf += 1;
	}

	return NULL;
}

static unsigned char *resync_force(unsigned char *head, unsigned char *tail, int unit_size)
{
	int i,n;
	unsigned char *buf;

	buf = head;
	while( buf < (tail-188) ){
		if(buf[0] == 0x47){
			n = (tail - buf) / unit_size;
			if(n == 0){
				return buf;
			}
			for(i=1;i<n;i++){
				if(buf[unit_size*i] != 0x47){
					break;
				}
			}
			if(i == n){
				return buf;
			}
		}
		buf += 1;
	}

	return NULL;
}

static void extract_ts_header(TS_HEADER *dst, unsigned char *packet)
{
	dst->sync                         =  packet[0];
	dst->transport_error_indicator    = (packet[1] >> 7) & 0x01;
	dst->payload_unit_start_indicator = (packet[1] >> 6) & 0x01;
	dst->transport_priority           = (packet[1] >> 5) & 0x01;
	dst->pid = ((packet[1] & 0x1f) << 8) | packet[2];
	dst->transport_scrambling_control = (packet[3] >> 6) & 0x03;
	dst->adaptation_field_control     = (packet[3] >> 4) & 0x03;
	dst->continuity_counter           =  packet[3]       & 0x0f;
}

static int check_unit_invert(unsigned char *head, unsigned char *tail)
{
	unsigned char *buf;

	buf = tail-188;

	while(head <= buf){
		if(buf[0] == 0x47){
			return tail-buf;
		}
		buf -= 1;
	}

	return 0;
}


